/* comm_link.c */
#define FALSE 0
#define TRUE  1

#include <stdio.h> 
#include "q4.h"
#include "comm_link.h"
#include "proto.h"

extern int anerror;


void open_link (comm) 
    comm_block *comm; 
  { /* open_link */
    comm->done_flag = FALSE;
    if (q4_get_my_id() == DNAML_MASTER_ID) {
      comm->send_type = DNAML_WORK;
      comm->send_to   = DNAML_DISPATCHER_ID;
      }
    else { /* Slave */
      comm->send_type = DNAML_RESULT;
      comm->send_to   = DNAML_MERGER_ID;
      }
  } /* open_link */


void close_link (comm) 
    comm_block *comm; 
  { /* close_link */
    /* No code exists here so far. */
  } /* close_link */


#if defined(SUN)
void nap(sleep_time)
    long sleep_time;
  { /* nap */
    struct timeval timeout;

    timeout.tv_sec  = sleep_time / 1000000L;  /* in sec. */
    timeout.tv_usec = sleep_time % 1000000L;  /* in microsec. */
    if (select(0, (fd_set *) 0, (fd_set *) 0, (fd_set *) 0, &timeout) < 0) {
      fprintf(stderr, "Proc#%d: select error in nap\n", q4_get_my_id());
      anerror = TRUE;
      }
  } /* nap */
#endif

void write_comm_msg (comm, message) 
    comm_block *comm; 
    char *message; 
  { /* write_comm_msg */

    if (comm->done_flag) {
      q4_send(DNAML_DONE, comm->send_to, NULL, 0);
      }
    else {
#     ifdef P4DEBUG
        { int my_id;
        if ((my_id = q4_get_my_id()) == DNAML_MASTER_ID)
          fprintf(stderr, "\nmaster sending msg: %s\n", message);
        else
          fprintf(stderr, "\nslave %d sending msg: %s\n", my_id, message);
        }
#     endif
      q4_send(comm->send_type, comm->send_to, message, strlen(message)+1);
      }
  } /* write_comm_msg */ 
 

void read_comm_msg(comm, message)
    comm_block *comm; 
    char *message; 
  { /* read_comm_msg */
    int my_id, type, from, size;
    char *msg;

    if ((my_id = q4_get_my_id()) == DNAML_MASTER_ID) {
      type = DNAML_RESULT;
      from = -1;
      }
    else  {
      type = -1;           /* DNAML_WORK || DNAML_DONE */
      from = DNAML_DISPATCHER_ID;
      }
#   if defined(SUN)
      while (! q4_messages_available(&type, &from))
        nap(20000);        /* sleep 20 millisec. */
#   endif
    msg  = NULL;
    q4_recv(&type, &from, &msg, &size);
    if (type == DNAML_DONE)
      comm->done_flag = TRUE;
    else if (type == DNAML_WORK || type == DNAML_RESULT) {
      comm->recv_from = from;
      strcpy(message, msg);
      q4_msg_free(msg);
#     ifdef P4DEBUG
        if (my_id == DNAML_MASTER_ID)
          fprintf(stderr, "\nmaster received msg: %s\n", message);
        else
          fprintf(stderr, "\nslave %d received msg: %s\n", my_id, message);
#     endif
      }
    else {
      if (my_id == DNAML_MASTER_ID)
        fprintf(stderr, "** master received unexpected type: %d\n", type);
      else
        fprintf(stderr, "** slave %d received unexpected type: %d\n", my_id,
                type);
      exit(2);
      }
  } /* read_comm_msg */
