/* dispatch.c */
#define FALSE 0
#define TRUE  1

#include <stdio.h>
#include "q4.h"
#include "comm_link.h"

#include "proto.h"

#define MAX_WORK_QUEUE_SIZE  5000
#define MAX_READY_QUEUE_SIZE 5000

/* typedef  int  boolean; */

typedef struct ready_queue { 
    int next;
    int length;
    int id[MAX_READY_QUEUE_SIZE];
} ready_queue; 

typedef struct work_queue { 
    int next;
    int length;
    char *work[MAX_WORK_QUEUE_SIZE];
    int size[MAX_WORK_QUEUE_SIZE];
} work_queue; 


#ifdef P4DEBUG
#  define  OUTFILE  Dspf
   FILE *OUTFILE;

#  define OUT_REQUEST(from) \
          do {\
	    fprintf(OUTFILE,"dispatcher received REQUEST from %d\n",from);\
            fflush(OUTFILE); \
          } while(0)
#  define OUT_SEND_WORK(to) \
               do {\
	    fprintf(OUTFILE,"dispatcher sends WORK to %d\n",to);\
            fflush(OUTFILE); \
	       } while(0)
#  define OUT_RECV_WORK \
               do {\
	    fprintf(OUTFILE,"dispatcher received WORK from MASTER\n");\
            fflush(OUTFILE); \
	       } while(0)
#  define OUT_RECV_DONE \
               do {\
	    fprintf(OUTFILE,"dispatcher received DONE from MASTER\n");\
            fflush(OUTFILE); \
	       } while(0)
#  define OUT_SEND_DONE(to) \
               do {\
	    fprintf(OUTFILE,"dispatcher sends DONE to %d\n",to);\
            fflush(OUTFILE); \
	       } while(0)
#else
#  define OUT_REQUEST(from)
#  define OUT_SEND_WORK(to)
#  define OUT_RECV_WORK
#  define OUT_RECV_DONE
#  define OUT_SEND_DONE(to)
#endif /* P4DEBUG */

ready_queue rd_q;
work_queue wk_q;


void init_ready_queue(Queue)
    ready_queue *Queue;
  { /* init_ready_queue */
    int i;

    Queue->next = 0;
    Queue->length = 0;
  } /* init_ready_queue */


void init_work_queue(Queue)
    work_queue *Queue;
  { /* init_work_queue */
    int i;

    Queue->next = 0;
    Queue->length = 0;
  } /* init_work_queue */


void put_ready_queue(Queue, id)
    ready_queue *Queue;
    int id;
  { /* put_ready_queue */
    int i;

    i = (Queue->next + Queue->length) % MAX_READY_QUEUE_SIZE;
    if (++(Queue->length) > MAX_READY_QUEUE_SIZE)
	fprintf(stderr, "put_ready_queue: ready queue overflow !!\n");
    else
	Queue->id[i] = id;
  } /* put_ready_queue */


void put_work_queue(Queue, wk, sz)
    work_queue *Queue;
    char *wk;
    int sz;
  { /* put_work_queue */
    int i;

    i = (Queue->next + Queue->length) % MAX_WORK_QUEUE_SIZE;
    if (++(Queue->length) > MAX_WORK_QUEUE_SIZE)
	fprintf(stderr, "put_work_queue: work queue overflow !!\n");
    else {
	Queue->work[i] = wk;
	Queue->size[i] = sz;
    }
  } /* put_work_queue */


boolean get_ready_queue(Queue, id)
    ready_queue *Queue;
    int *id;
  { /* get_ready_queue */
    if (Queue->length <= 0)
	return(FALSE);
    else {
	*id = Queue->id[Queue->next];
	Queue->next = (++Queue->next) % MAX_READY_QUEUE_SIZE;
	(Queue->length)--;
	return(TRUE);
        }
  } /* get_ready_queue */


boolean get_work_queue(Queue, wk, sz)
    work_queue *Queue;
    char **wk;
    int *sz;
  { /* get_work_queue */
    if (Queue->length <= 0)
	return(FALSE);
    else {
	*wk = Queue->work[Queue->next];
	*sz = Queue->size[Queue->next];
	Queue->next = (++Queue->next) % MAX_WORK_QUEUE_SIZE;
	(Queue->length)--;
	return(TRUE);
        }
  }  /* get_work_queue */


#if defined(P4DEBUG) && defined(INDEPENDENT)
FILE *fopen_pid (filenm, mode, name_pid)
    char *filenm, *mode, *name_pid;
  { /* fopen_pid */

    (void) sprintf(name_pid, "%s.%d", filenm, getpid());
    return  fopen(name_pid, mode);
  } /* fopen_pid */
#endif


dispatchslave()
  { /* dispatcher main */
    int     nprocs, type, sz, from, ready_slave, i;
    char    filename[128];
    char   *msg, *wk;
    boolean done;

    nprocs = q4_num_total_slaves();

#   ifdef P4DEBUG
      if ((OUTFILE = fopen_pid("disp.out", "w", filename)) == NULL) {
        fprintf(stderr,"dispatcher could not open output %s\n", filename);
        exit(1);
        }
#   endif

    init_ready_queue(&rd_q);
    init_work_queue(&wk_q);
    done = FALSE;
    while(!done) {
      type = from = -1;
      msg = NULL;
      q4_recv(&type, &from, &msg, &sz);
#ifdef P4DEBUG
      fprintf(OUTFILE, "dispatch received message type %d (size %d) from %d\n",
	      type,sz,from);
      fflush(OUTFILE);
#endif
      switch(type) {
        case DNAML_REQUEST:
          OUT_REQUEST(from);
          ready_slave = from;
          if (get_work_queue(&wk_q, &wk, &sz)) {
            OUT_SEND_WORK(ready_slave);
            q4_send(DNAML_WORK, ready_slave, wk, sz);
            q4_free(wk);
            }
          else
            put_ready_queue(&rd_q, ready_slave);
          break;
        case DNAML_WORK:
          OUT_RECV_WORK;
          if (get_ready_queue(&rd_q, &ready_slave)) {
            OUT_SEND_WORK(ready_slave);
            q4_send(DNAML_WORK, ready_slave, msg, sz);
            }
          else {
            wk = q4_malloc(strlen(msg)+1);   /* save work */
            strcpy(wk, msg);
            put_work_queue(&wk_q, wk, sz);
            }
          break;
        case DNAML_DONE:
          OUT_RECV_DONE;
          done = TRUE;
          for (i = DNAML_SLAVE_ID; i <= nprocs; i++) {
            OUT_SEND_DONE(i);
            q4_send(DNAML_DONE, i, NULL, 0);
            }
          OUT_SEND_DONE(DNAML_MERGER_ID);    /* notify done to merger */
          q4_send(DNAML_DONE, DNAML_MERGER_ID, NULL, 0);
          break;
        default:
          fprintf(stderr,"dispatcher got unexpected type message\n");
#         ifdef P4DEBUG
           fclose(OUTFILE);
#         endif
          exit(1);
        }
      q4_msg_free(msg);
      }
#   ifdef P4DEBUG
      fclose(OUTFILE);
#   endif
    q4_send(DNAML_DONE, DNAML_HOST_ID, NULL, 0);
  } /* slave */
