/* host.c */

#include <stdlib.h>
#include <stdio.h>
#include <math.h>
#include <sys/types.h>
#include <signal.h>
#include <unistd.h>

#define FALSE 0
#define TRUE  1

#include "q4.h"
#include "comm_link.h"
#include "proto.h"

/*  Global variables  */

#define  INFILE  Seqf
#define  OUTFILE Outf
FILE *INFILE, *OUTFILE;


/*  PROGRAM  */

#ifdef INDEPENDENT
FILE *freopen_pid (filenm, mode, stream)
    char *filenm, *mode;
    FILE *stream;
  { /* freopen_pid */
    char scr[512];

    (void) sprintf(scr, "%s.%d", filenm, getpid());
    return  freopen(scr, mode, stream);
  } /* freopen_pid */
#endif

void hostmain(argc,argv)
    int argc;
    char *argv[];
  { /* main */
    int nprocs, i, done, type, from, sz, nextsp;
    char *msg;
    long step_time;

#   ifdef DNAML_STEP
      int  dnaml_ln, dnaml_cur;
#   endif

    q4_initenv(&argc,argv);
    q4_create_procgroup();
    nprocs = q4_num_total_slaves();

    if (argc<2) {
	printf("Missing input file name\n");
#ifdef TANMOY
	kill(getpid(), SIGINT);  /* emulate terminal interrupt */
#else
	raise (SIGINT);
#endif
	exit(1);
        }
    if ((INFILE = fopen(argv[1], "r")) == NULL) {
	fprintf(stderr, "host could not open input file %s\n", argv[1]);
#ifdef TANMOY
	kill(getpid(), SIGINT);  /* emulate terminal interrupt */
#else
	raise (SIGINT);
#endif
	exit(1);
        }
    fclose(INFILE);

    /* Send input file name to master and all slaves */
    q4_send(DNAML_FILE_NAME, DNAML_MASTER_ID, argv[1], strlen(argv[1])+1);
    for (i = DNAML_SLAVE_ID; i <= nprocs; i++)
        q4_send(DNAML_FILE_NAME, i, argv[1], strlen(argv[1])+1);

#   ifdef DNAML_STEP
      if ((OUTFILE = freopen_pid("host.out", "w", stdout)) == NULL) {
        fprintf(stderr, "Could not open output file\n");
#ifdef TANMOY
	kill(getpid(), SIGINT);  /* emulate terminal interrupt */
#else
	raise (SIGINT);
#endif
        exit(1);
        }
      fprintf(OUTFILE, "Number of ids = %d\n", nprocs+1);

      type = DNAML_STEP_TIME;
      from = DNAML_MASTER_ID;
      msg = NULL;
      q4_recv(&type, &from, &msg, &sz);
      sscanf(msg, "%ld", &step_time);
      fprintf(stderr, "\n=== Step time = %10ld msec", step_time);
      fprintf(OUTFILE, "file input : %10ld msec\n", step_time);
      fflush(OUTFILE);
      q4_msg_free(msg);
      dnaml_ln = dnaml_cur = 1;
      fprintf(stderr, "\n   1\n   2\n   3 (%5d) ", 
              DNAML_CHAR_COUNT * (dnaml_ln - 1));
      nextsp = 4;
#   endif

    done = FALSE;
    while(!done) {
	type = from = -1;
	msg = NULL;
	q4_recv(&type,&from,&msg,&sz);
	switch(type) {
	  case DNAML_ADD_SPECS:
#           ifdef DNAML_STEP
              dnaml_ln = dnaml_cur = 1;
	      fprintf(stderr, "\n%4d (%5d) ", nextsp++, 
                      DNAML_CHAR_COUNT * (dnaml_ln - 1));
#           endif
	    break;
	  case DNAML_SEND_TREE:
#           ifdef DNAML_STEP
	      fputc('+', stderr);
	      dnaml_cur++;
#           endif
	    break;
	  case DNAML_RECV_TREE:
#           ifdef DNAML_STEP
	      fputc('-', stderr);
	      dnaml_cur++;
#           endif
	    break;
	  case DNAML_STEP_TIME:
	    sscanf(msg, "%ld", &step_time);
#           ifdef DNAML_STEP
              fprintf(OUTFILE, "~%4d Specs: %10ld msec\n", nextsp-1, 
                      step_time);
              fflush(OUTFILE);
	      fprintf(stderr, "\n=== Step time = %10ld msec", step_time);
#           endif
	    break;
	  case DNAML_DONE:
#           ifdef P4DEBUG
	       fprintf(stderr, "host got done from %d\n", from);
#           endif
	    nprocs--;
	    if (nprocs == 0)
		done = TRUE;
	    break;
          default:
	    fprintf(stderr, "What is this message %d from %d\n", type, from);
	    break;
	}
	q4_msg_free(msg);
#       ifdef DNAML_STEP
           if (dnaml_cur % (DNAML_CHAR_COUNT + 1) == 0) {
	     dnaml_ln++;
	     dnaml_cur = 1;
	     fprintf(stderr, "\n     (%5d) ", DNAML_CHAR_COUNT * (dnaml_ln-1));
	   }
#       endif
    }
#   ifdef DNAML_STEP
      fputc('\n', stderr);
      fclose(OUTFILE);
#   endif

    q4_wait_for_end();
  } /* main */
