#include <stdio.h>
#include <math.h>
#include <stdlib.h>
#include <string.h>
void 
tred2 (double *a, const int n, const int np, double *d, double *e)
{
#define a(i,j) a[(j-1)*np + (i-1)]
#define e(i)   e[i-1]
#define d(i)   d[i-1]
  int i, j, k, l;
  double f, g, h, hh, scale;
  for (i = n; i > 1; i--) {
    l = i-1;
    h = 0;
    scale = 0;
    if ( l > 1 ) {
      for ( k = 1; k <= l; k++ )
	scale += fabs(a(i,k));
      if (scale == 0) 
	e(i) = a(i,l);
      else {
	for (k = 1; k <= l; k++) {
	  a(i,k) /= scale;
	  h += a(i,k) * a(i,k);
	}
	f = a(i,l);
	g = -sqrt(h);
	if (f < 0) g = -g;
	e(i) = scale *g;
	h -= f*g;
	a(i,l) = f-g;
	f = 0;
	for (j = 1; j <=l ; j++) {
	  a(j,i) = a(i,j) / h;
	  g = 0;
	  for (k = 1; k <= j; k++)
	    g += a(j,k)*a(i,k);
	  for (k = j+1; k <= l; k++)
	    g += a(k,j)*a(i,k);
	  e(j) = g/h;
	  f += e(j)*a(i,j);
	}
	hh = f/(h+h);
	for (j = 1; j <= l; j++) {
	  f = a(i,j);
	  g = e(j) - hh * f;
	  e(j) = g;
	  for (k = 1; k <= j; k++) 
	    a(j,k) -= f*e(k) + g*a(i,k);
	}
      }
    } else 
      e(i) = a(i,l);
    d(i) = h;
  }
  d(1) = 0;
  e(1) = 0;
  for (i = 1; i <= n; i++) {
    l = i-1;
    if (d(i) != 0) {
      for (j = 1; j <=l; j++) {
	g = 0;
	for (k = 1; k <= l; k++)
	  g += a(i,k)*a(k,j);
	for (k=1; k <=l; k++)
	  a(k,j) -= g * a(k,i);
      }
    }
    d(i) = a(i,i);
    a(i,i) = 1;
    for (j=1; j<=l; j++)
      a(i,j) = a(j,i) = 0;
  }

  return;
#undef a
#undef e
#undef d
}

double pythag(double a, double b) {
  double absa = fabs(a), absb = fabs(b);
  return (absa > absb) ?
       absa * sqrt(1+ (absb/absa)*(absb/absa)) :
    absb == 0 ?
       0 :
       absb * sqrt(1+ (absa/absb)*(absa/absb));
}

void tqli(double *d, double *e, int n, int np, double *z) 
{
#define z(i,j) z[(j-1)*np + (i-1)]
#define e(i)   e[i-1]
#define d(i)   d[i-1]

int i, iter, k, l, m;
double b, c, dd, f, g, p, r, s;

for(i=2; i<=n; i++)
  e(i-1) = e(i);
e(n) = 0;

for (l = 1; l <= n; l++) {
  iter = 0;
label:
  for (m = l; (m < n); m++) {
    dd = fabs(d(m))+fabs(d(m+1));
    if (fabs(e(m))+dd == dd) break;
  }
  if (m != l) {
    if (iter == 30) {
      fprintf(stderr,"Too many iterations in tqli\n");
      exit(EXIT_FAILURE);
    }
    iter++;
    g = (d(l+1)-d(l))/(2*e(l));
    r = pythag(g,1.);
    g = d(m)-d(l)+e(l)/(g+(g<0?-r:r));
    s = 1; 
    c = 1;
    p = 0;
    for (i = m-1; i>=l; i--) {
      f = s*e(i);
      b = c*e(i);
      r = pythag(f,g);
      e(i+1) = r;
      if (r == 0) {
	d (i+1) -= p;
	e (m) = 0;
	goto label;
      }
      s = f/r;
      c = g/r;
      g = d(i+1)-p;
      r = (d(i)-g)*s + 2*c*b;
      p = s*r;
      d(i+1) = g + p;
      g = c*r - b;
      for (k=1; k <= n; k++) {
	f = z(k,i+1);
	z(k,i+1) = s * z(k,i) + c*f;
	z(k,i) = c * z(k,i) - s*f;
      }
    }
    d(l) -= p;
    e(l) = g;
    e(m) = 0;
    goto label;
  }
}

return;
#undef z
#undef e
#undef d

}
      
    
int main(void) {
  int i,m;
  FILE *infile = fopen("ratemat.in","r");
  char kw[30];
  char model[30] = ""; int numrates = 0;
  const char enumorder[] = "AGCT";
  double a [sizeof enumorder - 1] [sizeof enumorder - 1];
  double d[sizeof enumorder - 1], e[sizeof enumorder - 1];
  int order[sizeof enumorder - 1] = {0,1,2,3};
  double freqs[sizeof enumorder - 1], rates[12];
  double defaultfreqs[sizeof enumorder - 1] = {0};

  memcpy(freqs, defaultfreqs, sizeof freqs);

  if (!infile)
    return EXIT_SUCCESS;

  while(fscanf(infile,"%29s",kw) == 1) {
    if (strcmp(kw,"model:") == 0) {
      if(fscanf(infile,"%29s",model) != 1) {
	fprintf(stderr,"Can't parse model; Ignoring %s\n", "ratemat.in");
	return EXIT_FAILURE;
      }
      if (strcmp(model,"REV") == 0) {
	numrates = 5;
      } else {
	fprintf(stderr,"Only model: REV currently allowed; Ignoring %s\n",
		"ratemat.in");
	return EXIT_FAILURE;
      }
    } else if (strcmp(kw,"order:") == 0) {
      int k; char ch; char *find;
      for (k=0; k<sizeof enumorder - 1; k++) {
	if (fscanf(infile," %c", &ch) != 1) {
	  fprintf(stderr,"Can't parse order; Ignoring %s\n", "ratemat.in");
	  return EXIT_FAILURE;
	}
	find = strchr(enumorder,ch);
	if (!find || (find-enumorder >= sizeof enumorder - 1)) {
	  fprintf(stderr,"Can't parse order; Ignoring %s\n", "ratemat.in");
	  return EXIT_FAILURE;
	} else {
	  int l;
	  order[k] = find-enumorder;
	  for (l=0; l<k; l++) 
	    if(order[l] == order[k]) {
	      fprintf(stderr,"Can't parse order; Ignoring %s\n", "ratemat.in");
	      return EXIT_FAILURE;
	    }
	}
      }
    } else if (strcmp(kw,"freq:") == 0) {
      int k;
      for (k=0; k<sizeof enumorder - 1; k++)
	if (fscanf(infile,"%lf", &freqs[k]) != 1) {
	  fprintf(stderr,"Can't parse freq; Ignoring %s\n", "ratemat.in");
	  return EXIT_FAILURE;
	}
    } else if (strcmp(kw,"ratepar:") == 0) {
      int k;
      for (k=0; k<numrates; k++)
	if (fscanf(infile,"%lf", &rates[k]) != 1) {
	  fprintf(stderr,"Can't parse ratepar; Ignoring %s\n", "ratemat.in");
	  return EXIT_FAILURE;
	}
    } else {
      fprintf(stderr,"Invalid kw: %s; Ignoring %s\n",kw, "ratemat.in");
      return EXIT_FAILURE;
    }
  }

  m = 0;
  for(i=0; i<sizeof enumorder - 1; i++) 
    a[i][i] = 0;
  for(i=0; i<sizeof enumorder - 1; i++) {
    int j;
    for(j=i+1; j<sizeof enumorder - 1; j++) {
      double factor = ((m>=numrates) ? 1 : rates[m++]);
      a[order[i]][order[j]] = a[order[j]][order[i]] = 
	factor * sqrt( freqs[i] * freqs[j]);
      a[order[i]][order[i]] -= factor * freqs[j];
      a[order[j]][order[j]] -= factor * freqs[i];
    }
  }
  
  printf("order: ");
  for(i=0; i<sizeof enumorder - 1; i++)
    printf("%d ",order[i]);
  printf("\n");

  printf("freqs: ");
  for(i=0; i<sizeof enumorder - 1; i++)
    printf("%f ",freqs[i]);
  printf("\n");
 
  printf("rates: ");
  for(i=0; i<numrates; i++)
    printf("%f ",rates[i]);
  printf("\n");

  printf("\nmatrix:\n");
  for(i=0; i<sizeof enumorder - 1; i++) {
    int j;
    for(j=0; j<sizeof enumorder - 1; j++) {
      printf("%f ", a[i][j]);
    }
    printf("\n");
  }
  printf("\n");

  tred2((double *)a,sizeof enumorder - 1,sizeof enumorder - 1,d,e);
  tqli(d, e, sizeof enumorder - 1, sizeof enumorder - 1, (double*) a);

  for(i=0; i<sizeof enumorder - 1; i++) {
    int j;
    for(j=0; j<sizeof enumorder - 1; j++) {
      a[order[i]][order[j]] *= sqrt(freqs[j]); 
    }
  }

  for (i=0; i<sizeof enumorder - 1; i++)
    if (d[i] > -1e-8) {
      if (i != 0) {
	int j;
	double tmp = d[i], sum=0;
	d[i] = d[0];
	d[0] = tmp;
	for (j=0; j < sizeof enumorder - 1; j++) {
	  tmp = a[i][j];
	  a[i][j] = a[0][j];
	  sum += (a[0][j] = tmp);
	}
	for (j=0; j < sizeof enumorder - 1; j++) a[0][j] /= sum;
      }
      break;
    }

  printf("Rates:\n");
  printf("%f %f %f %f\n", (d[0]), (d[1]), (d[2]), (d[3]));

  printf("Directions:\n");
  for(i=0; i<sizeof enumorder - 1; i++)
    printf("%f %f %f %f\n", a[i][0], a[i][1], a[i][2], a[i][3]);
  printf("\n");
  
  return 0;
}

