 /*  merge.c */

#include <stdio.h>
#include "ml.h"
#include "q4.h"
#include "comm_link.h" 

#include "proto.h"

boolean anerror;

#ifdef DNAML_STEP
#  define REPORT_RECV_TREE q4_send(DNAML_RECV_TREE, DNAML_HOST_ID, NULL, 0)
#else
#  define REPORT_RECV_TREE
#endif

#ifdef P4DEBUG
#  define  OUTFILE  Mergf
   FILE *OUTFILE;
#  define OUT_NUM_TREE(num) \
     do {\
      fprintf(OUTFILE, "merger received n_tree_sent = %d from master\n", num);\
      fflush(OUTFILE);\
     } while(0)
#  define OUT_RECV_TREE(from) \
     do {\
      fprintf(OUTFILE, "merger received a tree from %d\n", from);\
      fflush(OUTFILE);\
								   } while(0)
#  define OUT_SEND_TREE \
     do {\
      fprintf(OUTFILE, "merger sends the best tree to master\n");\
      fflush(OUTFILE);\
								   } while(0)
#  define OUT_RECV_DONE \
     do {\
      fprintf(OUTFILE, "merger received DONE from dispatcher\n");\
      fflush(OUTFILE);\
								   } while(0)
#else
#  define OUT_NUM_TREE(num)
#  define OUT_RECV_TREE(from)
#  define OUT_SEND_TREE
#  define OUT_RECV_DONE
#endif

#define CHECK_TO_SEND \
{ \
    if (n_tree_sent && (n_tree_sent <= n_tree_received)) {\
      OUT_SEND_TREE; \
      q4_send(DNAML_RESULT, DNAML_MASTER_ID, treestr, strlen(treestr)+1); \
      n_tree_received -= n_tree_sent;  n_tree_sent = 0; \
      best_likelihood = unlikely; \
      } \
}

#if INDEPENDENT

FILE *fopen_pid (filenm, mode, name_pid)
    char *filenm, *mode, *name_pid;
  { /* fopen_pid */

    (void) sprintf(name_pid, "%s.%d", filenm, getpid());
    return  fopen(name_pid, mode);
  } /* fopen_pid */


char * strchr (str, chr)
    char *str;
    int   chr;
 { /* strchr */
    int  c;

    while (c = *str)  {if (c == chr) return str; str++;}
    return  (char *) NULL;
 } /* strchr */


char * strstr (str1, str2)
    char *str1, *str2;
 { /* strstr */
    char *s1, *s2;
    int  c;

    while (*(s1 = str1)) {
      s2 = str2;
      do {
        if (! (c = *s2++))  return str1;
        } 
        while (*s1++ == c);
      str1++;
      }
    return  (char *) NULL;
 } /* strstr */


boolean readKeyValue (string, key, format, value)
    char *string, *key, *format;
    void *value;
  { /* readKeyValue */

    if (! (string = strstr(string, key)))  return FALSE;
    string += strlen(key);
    if (! (string = strchr(string, '=')))  return FALSE;
    string++;
    return  sscanf(string, format, value);  /* 1 if read, otherwise 0 */
  } /* readKeyValue */


double str_readTreeLikelihood (treestr)
    char *treestr;
  { /* str_readTreeLikelihood */
    double lk1;
    char    *com, *com_end;
    boolean  readKeyValue();

    if ((com = strchr(treestr, '[')) && (com < strchr(treestr, '('))
                                     && (com_end = strchr(com, ']'))) {
      com++;
      *com_end = 0;
      if (readKeyValue(com, likelihood_key, "%lg", (void *) &(lk1))) {
        *com_end = ']';
        return lk1;
        }
      }

    fprintf(stderr, "ERROR reading likelihood in receiveTree\n");
    /* THIS WAS return badEval; */
    anerror = TRUE;
    return 1.0;
  } /* str_readTreeLikelihood */

#endif

int
mergeslave ()
  { /* slave */
    int     type, sz, from;
    int     n_tree_sent, n_tree_received;
    double  best_likelihood, got_likelihood;
    char    filename[128];
#define maxsp 300
    char   *msg, *treestrp, treestr[maxsp*(nmlngth+120)+1024];
    boolean done;

#   ifdef P4DEBUG
      if ((OUTFILE = fopen_pid("merge.out", "w", filename)) == NULL) {
        fprintf(stderr,"merger could not open output %s\n", filename);
        exit(1);
        }
#   endif

    anerror = FALSE;
    done = FALSE;
    n_tree_sent = n_tree_received = 0;
    best_likelihood = unlikely;
    while(!done) {
      type = from = -1;
      msg = NULL;
      q4_recv(&type, &from, &msg, &sz);
#ifdef P4DEBUG
      fprintf(OUTFILE, "merger received message type %d (size %d) from %d\n",
	      type,sz,from);
      fflush(OUTFILE);
#endif
      switch(type) {
        case DNAML_NUM_TREE:
          if (sscanf(msg, "%d", & n_tree_sent) != 1) {
            fprintf(stderr, "ERROR reading n_tree_sent in merger\n");
            exit(1);
            }
          OUT_NUM_TREE(n_tree_sent);
          CHECK_TO_SEND;
          break;
        case DNAML_RESULT:
          OUT_RECV_TREE(from);
          REPORT_RECV_TREE;
          n_tree_received++;
          if ((got_likelihood=str_readTreeLikelihood(msg)) == badEval) {
            if (anerror) exit(1);
	  }
	  else if (got_likelihood > best_likelihood) {
            if (anerror) exit(1);
            best_likelihood = got_likelihood;
            strcpy(treestr,msg);
	  }
	  else if (got_likelihood == badEval) {
	    if (anerror) exit(1);
	  }
          CHECK_TO_SEND;
          break;
        case DNAML_DONE:  /* dispatcher sends done to merger */
          OUT_RECV_DONE;
          done = TRUE;
          break;
        default:
          fprintf(stderr, "merger got unexpected type message\n");
#         ifdef P4DEBUG
            fclose(OUTFILE);
		    exit(1);
#         endif
          }
      q4_msg_free(msg);
      }
#   ifdef P4DEBUG
      fclose(OUTFILE);
#   endif
    q4_send(DNAML_DONE, DNAML_HOST_ID, NULL, 0);

    return 0;
  } /* slave */
