boolean newview (tr, p)                      /*  Update likelihoods at node */
    tree    *tr;
    nodeptr  p;
  { /* newview */
    if (p->tip) {

      /*  Make sure that data are at tip */

      int i;
      yType *yptr;
      xtype (*x3)[4];
      double (*EIGV)[4][4] = tr->rdta->EIGV;
      int  *cptr = tr->cdta->patcat;

#ifdef DOCOV
      if (p->x) /* return TRUE; */ /*  They are already there */ 
	/* Tanmoy: Update anyway */; else
#else
      if (p->x) return TRUE;  /*  They are already there */
#endif
      if (! getxtip(p)) return FALSE; /*  They are not, so get memory */

      x3 = p->x->u; yptr = p->tip;
      for (i = 0; i < tr->cdta->endsite; i++) {
	int j;
        int code = *yptr++;
	int cat = *cptr++;

	for (j = 0; j < 4; j++)
	  (*x3)[j] = 0;

	for (j = 0; j < 4; j++) {
	  if ((code >> j) & 1) {
	    int k;
	    int jj = "0213"[j] - '0';  /* The matrix is in agct notation, 
					  the bits   are in acgt!!! */
	    for (k = 0; k < 4; k++)
	      (*x3)[k] += EIGV[cat][jj][k];
	  }
	}
	x3++;
      }
      return TRUE;

    }  else {

      /*  Internal node needs update */

      int  i;
      xtype   (*x1)[4], (*x2)[4], (*x3)[4];
      double   z1, lz1, z2, lz2;
      nodeptr  q, r;

      q = p->next->back;
      r = p->next->next->back;

      while ((! p->x) || (! q->x) || (! r->x)) {
	if (! q->x) if (! newview(tr, q))  {
	  return FALSE;
	}
	if (! r->x) if (! newview(tr, r))  {
	  return FALSE;
	}
	if (! p->x) if (! getxnode(p)) {
	  return FALSE;
	}
      }

      x1  = q->x->u;
      z1  = q->z;
      lz1 = (z1 > zmin) ? log(z1) : log(zmin);

      x2  = r->x->u;
      z2  = r->z;
      lz2 = (z2 > zmin) ? log(z2) : log(zmin);

      x3  = p->x->u;
      { 
	/* newview not called in this block: so make the big stuff static;
	   stupid compiler allocates this at the beginning. */
	static double  diagptable[maxcategories+1][4][2];

	double (*diagp)[4][2] = &diagptable[1];
	int  *cptr = tr->cdta->patcat;
	double (*EIGN)[4]    = &(tr->rdta->EIGN[1]);
	double (*invfreq)[4] = tr->rdta->invfreq;
	double (*EIGV)[4][4] = tr->rdta->EIGV;

	for (i = 1; i <= tr->rdta->categs; i++) {
	  /* exps for each category */
	  int j;
	  
	  for (j = 1; j < 4; j++) {
	    (*diagp)[j][0] = exp ((*EIGN) [j] * lz1);
	    (*diagp)[j][1] = exp ((*EIGN) [j] * lz2);
	  }
	  diagp++; EIGN++;
	}

	for (i = 0; i < tr->cdta->endsite; i++) {
	  int j; 
	  xtype tmp[4][2], ump[4][2];
	  int cat = *cptr++;

	  /* multiply by eigenvalue */
	  for (j = 1; j < 4; j++) {
	    tmp[j][0] = (*x1)[j] * diagptable[cat][j][0];
	    tmp[j][1] = (*x2)[j] * diagptable[cat][j][1];
	  }

	  /* multiply by inverse eigenvector */
	  for (j = 0; j < 4; j++) {
	    int k;
	    ump[j][0] = ump[j][1] = 0;
	    for (k = 1; k < 4; k++) {
	      ump[j][0] += tmp[k][0] * EIGV[cat][j][k];
	      ump[j][1] += tmp[k][1] * EIGV[cat][j][k];
	    }
	    ump[j][0] *= invfreq[cat][j];
	    ump[j][0] += (*x1)[0];
	    ump[j][1] *= invfreq[cat][j];
	    ump[j][1] += (*x2)[0];
	  }

	  /* multiply and by eigenvector */
	  for (j = 0; j < 4; j++) {
	    int k;
	    (*x3)[j] = 0;
	    for (k = 0; k < 4; k++)
	      (*x3)[j] += ump[k][0] * ump[k][1] * EIGV[cat][k][j];
	  }
	  x1++; x2++; x3++;
	} /* sites */
      } /* block */
      return TRUE;
    } /* else */
  } /* newview */

double evaluate (tr, p)
    tree    *tr;
    nodeptr  p;
  { /* evaluate */
    double   sum, z, lz;
    double  *log_f = tr->log_f;
    nodeptr  q;
    int     i;
    int     *wptr = tr->cdta->aliaswgt;
    int     *cptr = tr->cdta->patcat;
    double  diagptable[maxcategories+1][4], (*diagp)[4] = &diagptable[1];
    double (*EIGN)[4]    = &(tr->rdta->EIGN[1]);
    xtype   (*x1)[4], (*x2)[4];


    q = p->back;
    while ((! p->x) || (! q->x)) {
      if (! (p->x)) if (! newview(tr, p)) return badEval;
      if (! (q->x)) if (! newview(tr, q)) return badEval;
    }


    x1  = p->x->u;
    x2  = q->x->u;
    z = p->z;
    if (z < zmin) z = zmin;
    lz = log(z);

    for (i = 1; i <= tr->rdta->categs; i++) {
      int j;
      (*diagp)[0] = 1;
      for (j = 1; j < 4; j++) {
	(*diagp)[j] = exp ((*EIGN) [j] * lz);
      }
      diagp++; EIGN++;
    }

    tr->log_f_valid = tr->cdta->endsite;
    sum = 0.0;

    for (i = 0; i < tr->cdta->endsite; i++) {
      int j;
      double term = 0;
      int    cat  = *cptr++;
      for (j = 0; j < 4; j++)
	term += (*x1)[j] * (*x2)[j] * diagptable[cat][j];
      term = log(term);
      sum += *wptr++ * term;
      *log_f++ = term;
      x1++; x2++;
    }

    tr->likelihood = sum;

    return  sum;
  } /* evaluate */


double makenewz (tr, p, q, z0, maxiter)
    tree    *tr;
    nodeptr  p, q;
    double   z0;
    int  maxiter;
  { /* makenewz */
    double   z, zprev, zstep;
    xtype   (*x1)[4], (*x2)[4];
    double  (*sumtable)[4], (*sum)[4];
    int     i;
    double  dlnLdlz = 0.0;                 /*  = d(ln(likelihood))/d(lz) */
    double  d2lnLdlz2 = 0.0;               /*  = d2(ln(likelihood))/d(lz)2 */

    while ((! p->x) || (! q->x)) {
      if (! (p->x)) if (! newview(tr, p)) return badZ;
      if (! (q->x)) if (! newview(tr, q)) return badZ;
    }

    x1 = p->x->u;
    x2 = q->x->u;

    {
      size_t scratch_size = sizeof(double[4]) * tr->cdta->endsite;
      if ((sum = sumtable = (double (*) [4]) Malloc(scratch_size)));
      else {
        printf("ERROR: makenewz unable to obtain space for arrays\n");
        return badZ;
      }
    }

    for (i = 0; i < tr->cdta->endsite; i++) {
      int j;
      for (j = 0; j < 4; j++)
	(*sum)[j] = (*x1)[j] * (*x2)[j];
      x1++; x2++; sum++;
    }

    z = z0;
    do {
      int curvatOK = FALSE;

      zprev = z;
      zstep = (1.0 - zmax) * z + zmin;

      do {
	double  diagptable[maxcategories+1][4], (*diagp)[4] = &diagptable[1];
	double (*EIGN)[4]    = &(tr->rdta->EIGN[1]);
	double lz;
        int    *cptr    = &(tr->cdta->patcat[0]);
        double *wrptr   = &(tr->cdta->wr[0]);
        double *wr2ptr  = &(tr->cdta->wr2[0]);

        dlnLdlz = 0.0;                 /*  = d(ln(likelihood))/d(lz) */
        d2lnLdlz2 = 0.0;               /*  = d2(ln(likelihood))/d(lz)2 */

        if (z < zmin) z = zmin;
        else if (z > zmax) z = zmax;
        lz    = log(z);

        for (i = 1; i <= tr->rdta->categs; i++) {
	  int j;
	  (*diagp)[0] = 1;
	  for (j = 1; j < 4; j++) {
	    (*diagp)[j] = exp ((*EIGN) [j] * lz);
	  }
	  diagp++; EIGN++;
	}

	EIGN = tr->rdta->EIGN;
	sum = sumtable;
        for (i = 0; i < tr->cdta->endsite; i++) {
	  double tmp[4];
          int    cat    = *cptr++;
	  double inv_Li = 0, dlnLidlz = 0, d2lnLidlz2 = 0;
	  int j;

	  for (j = 0; j < 4; j++) {
	    inv_Li += (tmp[j] = (*sum)[j] * diagptable[cat][j]);
	  }

          inv_Li = 1.0/inv_Li;

	  for (j = 1; j < 4; j++) {
	    double t = tmp[j] * EIGN[cat][j];
	    dlnLidlz   += t;
	    d2lnLidlz2 += t * EIGN[cat][j];
	  }

	  dlnLidlz   *= inv_Li;
	  d2lnLidlz2 *= inv_Li;
          dlnLdlz   += *wrptr++  * dlnLidlz;
          d2lnLdlz2 += *wr2ptr++ * (d2lnLidlz2 - dlnLidlz * dlnLidlz);
	  sum++;
	}

        if ((d2lnLdlz2 >= 0.0) && (z < zmax))
          zprev = z = 0.37 * z + 0.63;  /*  Bad curvature, shorten branch */
        else
          curvatOK = TRUE;

      } while (! curvatOK);

      if (d2lnLdlz2 < 0.0) {
	double tantmp = -dlnLdlz / d2lnLdlz2;  /* prevent overflow */
	if (tantmp < 100) {
	  z *= exp(tantmp);
	  if (z < zmin) z = zmin;
	  if (z > 0.25 * zprev + 0.75)    /*  Limit steps toward z = 1.0 */
	    z = 0.25 * zprev + 0.75;
	} else {
	  z = 0.25 * zprev + 0.75;
	}
      }
      if (z > zmax) z = zmax;

    } while ((--maxiter > 0) && (ABS(z - zprev) > zstep));

    Free (sumtable);

    return  z;
  } /* makenewz */

double sigma (tr, p, sumlrptr)
    tree    *tr;
    nodeptr  p;
    double  *sumlrptr;
    /* compute standard deviation */
  { /* sigma */
    double  slope, sum, sumlr, z, lz;
    nodeptr  q;
    int      i, *wptr = tr->cdta->aliaswgt;
    int     *cptr = tr->cdta->patcat;
    xtype   (*x1)[4], (*x2)[4];
    double (*EIGV)[4][4] = tr->rdta->EIGV;
    double (*EIGN)[4]    = tr->rdta->EIGN;
    double (*invfreq)[4] = tr->rdta->invfreq;

    q = p->back;
    while ((! p->x) || (! q->x)) {
      if (! (p->x)) if (! newview(tr, p)) return -1.0;
      if (! (q->x)) if (! newview(tr, q)) return -1.0;
    }

    x1 = p->x->u;
    x2 = q->x->u;

    z = p->z;
    if (z < zmin) z = zmin;
    lz = log(z);

    sum = sumlr = slope = 0.0;

    for (i = 0; i < tr->cdta->endsite; i++) {
      double suma=1e-300, d2=0, li=0, temp=0, w=*wptr++;
      int cat  = *cptr++;
      double diagp[4];
      int j;

      diagp[0] = 1;
      for (j = 1; j < 4; j++)
	diagp[j] = exp(EIGN[cat][j] *lz);

      for (j = 0; j < 4; j++) {
	double u = 0, t = (*x1)[j] * (*x2)[j];
	int k;

	/* The following is always 1: so delete this */
	for (k = 0; k < 4; k++) 
	  u += (EIGV[cat][k][j]*EIGV[cat][k][j]) * invfreq[cat][k];
	suma += t * u;

	t *= diagp[j];
	li   += t;
	temp += t * EIGN[cat][j];
	d2   += t * EIGN[cat][j] * (EIGN[cat][j] - 1);
      }
      temp /= li;

      slope += w *  temp;
      sum   += w * (temp * temp - d2/li);
      sumlr += w * log(li / suma);

      x1++; x2++;

    }

    *sumlrptr = sumlr;

    return (sum > 1.0E-300) ? z*(-slope + sqrt(slope*slope + 3.841*sum))/sum
                            : 1.0;
  } /* sigma */

#ifdef DOCOV
int namelinks(tree *tr, nodeptr p, int linkname) {

  nodeptr q = p->back->tip?p->back:p; /* take care of init: assumes 
					 start is a tip */

  q->linknumber = linkname;
  q->back->linknumber = linkname;
  printf("%10d: ",linkname++);
  
  printf("%4d          ", q->back->number - tr->mxtips);

  if (q->tip) {
    char ch, *nameptr; int k;
    nameptr = q->name;
    k = nmlngth;
    while ((ch = *nameptr++)) {putchar(ch); k--;}
    while (--k >= 0) putchar(' ');
  }
  else {
    int k;
    printf("%4d", q->number - tr->mxtips);
    for (k = 4; k < nmlngth; k++) putchar(' ');
  }

  putchar('\n');

  return p->tip ? linkname :
    namelinks(tr,p->next->next->back,
	      namelinks(tr,p->next->back,linkname));

}


int covdescribe(tree *tr, nodeptr rr, nodeptr pp, int l1, int l2) {
  nodeptr p = pp->back->tip?pp->back:pp;
  nodeptr q = p->back;
  nodeptr ss = ((pp==rr) || (pp->back==rr)) ? NULL : rr ;
  xtype (*x1)[4], (*x2)[4];
  double z = p->z < zmin ? zmin : p->z;
  double lz = log(z);
  int i, *wptr = tr->cdta->aliaswgt;
  int *cptr = tr->cdta->patcat;
  double (*EIGN)[4] = tr->rdta->EIGN;
  double slope = 0, cov = 0, slope2 = 0;
  double *log_f = tr->log_f;
  int l2tmp;
  int count = 0;
  static double *deriv1 = 0;

  if (!deriv1) deriv1 = malloc(tr->cdta->endsite * sizeof (double));

  newview2(tr,pp,rr); 
  newview2(tr,pp->back,ss);
  while (! (pp->x && pp->back->x) && count++ < 1000) {
    newview2(tr,pp,rr); 
    newview2(tr,pp->back,ss);
  }

  if (count >= 1000) return -1;

  if (tr->log_f_valid != tr->cdta->endsite) {
    printf ("WARNING: Impossible situation, Calling evaluate %d %d\n",l1,l2);
    evaluate(tr, pp);
  }


  newview2(tr,pp,rr); 
  newview2(tr,pp->back,ss);
  while (! (pp->x && pp->back->x) && count++ < 1000) {
    newview2(tr,pp,rr); 
    newview2(tr,pp->back,ss);
  }

  x1 = p->x->u;  x2 = q->x->u;

  for (i = 0; i < tr->cdta->endsite; i++) {
    double w  = *wptr++;
    int  cat  = *cptr++;
    double li = exp(*log_f++);
    double s1=0, s2=0;
    int j;

    for (j = 0; j < 4; j++) {
      double t = (*x1)[j] * (*x2)[j] * exp(EIGN[cat][j]*lz);

      s1 += t;
      s2 += t * EIGN[cat][j];

    }
    s1 /= li; s2 /= li;

    slope += w *  s1;
    cov   += w *  s2;
    deriv1[i] = s1;

    x1++; x2++;
  }

  newview2(tr,pp,NULL); 
  newview2(tr,pp->back,NULL);
  while (! (pp->x && pp->back->x) && count++ < 1000) {
    newview2(tr,pp,NULL); 
    newview2(tr,pp->back,NULL);
  }

  x1 = p->x->u;  x2 = q->x->u;
  log_f = tr->log_f;
  cptr = tr->cdta->patcat;
  wptr = tr->cdta->aliaswgt;

  for (i = 0; i < tr->cdta->endsite; i++) {
    double w  = *wptr++;
    int  cat  = *cptr++;
    double li = exp(*log_f++);
    double s2=0;
    int j;

    for (j = 0; j < 4; j++) {
      double t = (*x1)[j] * (*x2)[j] * exp(EIGN[cat][j]*lz);

      s2 += t * EIGN[cat][j];

    }
    s2 /= li;

    slope2 += w * s2;
    cov -= w * s2 * deriv1[i];

    x1++; x2++;
  }

  printf("%10d %10d: %20g %20g %20g\n",rr->linknumber,p->linknumber,
	 -cov/(tr->rdta->fracchangeav * tr->rdta->fracchangeav),
          slope/tr->rdta->fracchangeav,
          slope2/tr->rdta->fracchangeav);

  fflush(NULL);

  l2tmp = l2 % (2 * tr->mxtips - 3) + 1;
  if (! pp->tip ) {
    l2tmp = covdescribe(tr,rr,pp->next->back,l1,l2tmp);
    l2tmp = covdescribe(tr,rr,pp->next->next->back,l1,l2tmp);
  }

  if (l1 == l2) {
    if (! rr->back->tip) {
      l2tmp = covdescribe(tr,rr->back,rr->back->next->back,l1,l2tmp);
      l2tmp = covdescribe(tr,rr->back,rr->back->next->next->back,l1,l2tmp);
    }
    l1++;
    if (! rr->tip) {
      l1 = covdescribe(tr,rr->next->back,rr->next->back,l1,l1);
      l1 = covdescribe(tr,rr->next->next->back,rr->next->next->back,
			  l1,l1);
    }
    return l1;
  } else
    return l2tmp;
}

boolean newview2 (tr, p, qq)              /*  Update likelihoods at node */
    tree    *tr;
    nodeptr  p, qq;
  { /* newview */

    if (p->tip) {

      /*  Make sure that data are at tip */

      int i;
      yType *yptr;
      xtype (*x3)[4];
      double (*EIGV)[4][4] = tr->rdta->EIGV;
      double (*EIGN)[4]    = tr->rdta->EIGN;
      int  *cptr = tr->cdta->patcat;

      
      if (! p->x ) {
	p->x = (xarray *) Malloc(sizeof (xarray));
	p->x->u = (xtype (*) [4]) Malloc(tr->cdta->endsite * sizeof(xtype[4]));
	p->x->prev = p->x->next = p->x;
	p->x->owner = p;
#if 0	  
	if (! getxtip(p)) {
	  printf("NO TIP LEFT\n");
	  return FALSE; /*  They are not, so get memory */
	}
#endif
      }

      x3 = p->x->u; yptr = p->tip;
      for (i = 0; i < tr->cdta->endsite; i++) {
	int j;
        int code = *yptr++;
	int cat = *cptr++;

	for (j = 0; j < 4; j++)
	  (*x3)[j] = 0;

	for (j = 0; j < 4; j++) {
	  if ((code >> j) & 1) {
	    int k;
	    int jj = "0213"[j] - '0';  /* The matrix is in agct notation, 
					  the bits   are in acgt!!! */
	    for (k = 0; k < 4; k++)
	      (*x3)[k] += EIGV[cat][jj][k];
	  }
	}
	if ( (p == qq) || (p->back == qq)) 
	  for (j = 0; j < 4; j++)
	    (*x3)[j] *= EIGN[cat][j];

	x3++;
      }

      return TRUE;

    }  else {

      /*  Internal node needs update */

      int  i;
      xtype   (*x1)[4], (*x2)[4], (*x3)[4];
      double   z1, lz1, z2, lz2;
      nodeptr  q, r;
      int count = 0;

      q = p->next->back;
      r = p->next->next->back;

      newview2(tr, q, qq);
      newview2(tr, r, qq);
      while (! (q->x && r->x && p->x ) && count++ < 1000) {
	if (! q->x) newview2(tr, q, qq);
	if (! r->x) newview2(tr, r, qq);
	if (! p->x) if (! getxnode(p)) {
	  printf("NO NODE LEFT\n");
	  return FALSE;
	}
      }

      if (count >= 1000) return FALSE;

      x1  = q->x->u;
      z1  = q->z;
      lz1 = (z1 > zmin) ? log(z1) : log(zmin);

      x2  = r->x->u;
      z2  = r->z;
      lz2 = (z2 > zmin) ? log(z2) : log(zmin);

      x3  = p->x->u;
      { 
	/* newview not called in this block: so make the big stuff static;
	   stupid compiler allocates this at the beginning. */
	static double  diagptable[maxcategories+1][4][2];

	double (*diagp)[4][2] = &diagptable[1];
	int  *cptr = tr->cdta->patcat;
	double (*EIGN)[4]    = &(tr->rdta->EIGN[1]);
	double (*invfreq)[4] = tr->rdta->invfreq;
	double (*EIGV)[4][4] = tr->rdta->EIGV;

	for (i = 1; i <= tr->rdta->categs; i++) {
	  /* exps for each category */
	  int j;
	  
	  for (j = 1; j < 4; j++) {
	    (*diagp)[j][0] = exp ((*EIGN) [j] * lz1);
	    (*diagp)[j][1] = exp ((*EIGN) [j] * lz2);
	  }
	  diagp++; EIGN++;
	}

	EIGN = tr->rdta->EIGN;

	for (i = 0; i < tr->cdta->endsite; i++) {
	  int j; 
	  xtype tmp[4][2], ump[4][2];
	  int cat = *cptr++;

	  /* multiply by eigenvalue */
	  for (j = 1; j < 4; j++) {
	    tmp[j][0] = (*x1)[j] * diagptable[cat][j][0];
	    tmp[j][1] = (*x2)[j] * diagptable[cat][j][1];
	  }

	  /* multiply by inverse eigenvector */
	  for (j = 0; j < 4; j++) {
	    int k;
	    ump[j][0] = ump[j][1] = 0;
	    for (k = 1; k < 4; k++) {
	      ump[j][0] += tmp[k][0] * EIGV[cat][j][k];
	      ump[j][1] += tmp[k][1] * EIGV[cat][j][k];
	    }
	    ump[j][0] *= invfreq[cat][j];
	    ump[j][0] += (*x1)[0];
	    ump[j][1] *= invfreq[cat][j];
	    ump[j][1] += (*x2)[0];
	  }

	  /* multiply and by eigenvector */
	  for (j = 0; j < 4; j++) {
	    int k;
	    (*x3)[j] = 0;
	    for (k = 0; k < 4; k++)
	      (*x3)[j] += ump[k][0] * ump[k][1] * EIGV[cat][k][j];
	  }
	  if ( (p == qq) || (p->back == qq) ) 
	    for (j = 0; j < 4; j++)
	      (*x3)[j] *= EIGN[cat][j];
	  x1++; x2++; x3++;
	} /* sites */
      } /* block */
      return TRUE;
    } /* else */
  } /* newview */

#endif
