#include "q4.h"
#include <stdlib.h>

/* mpi.h ungood */
#ifdef __STDC__
#define USE_STDARG
#endif
#include <mpi.h>

#include <stdio.h>
#include <time.h>
#include "comm_link.h"
#include "proto.h"

/* Provide dummy implementations of these */
int q4_clock() {
  static time_t start;
  static int init;
  if (!init) {
    (void)time(&start);
    init++;
  }
  return 1000*difftime(time(NULL),start);
}
int q4_get_my_id() {
  int me;
  if (MPI_Comm_rank(MPI_COMM_WORLD,&me) != MPI_SUCCESS) {
    fprintf (stderr, "Failed to get me\n");
  }
  return me;
}
int q4_num_total_slaves () {
  int size;
  if (MPI_Comm_size(MPI_COMM_WORLD,&size) != MPI_SUCCESS) {
    fprintf (stderr, "Failed to get size\n");
  }
  return size-1;
}

void q4_send(type,to,message,size) 
int type, to, size;
const char *message;
{
  if(MPI_Send(message, size, MPI_CHAR, to, type, MPI_COMM_WORLD)
     != MPI_SUCCESS) {
    fprintf (stderr, "Send failed\n");
  }
}

void q4_recv(type,from,message,size)
int *type, *from, *size;
char **message;
{ 
  MPI_Status status;
  if (*type < 0) *type = MPI_ANY_TAG;
  if (*from < 0) *from = MPI_ANY_SOURCE;
  if (MPI_Probe(*from, *type, MPI_COMM_WORLD, &status) != MPI_SUCCESS) {
    fprintf(stderr,"probe failed\n");
  }
  if (MPI_Get_count(&status, MPI_CHAR, size) != MPI_SUCCESS ||
      *size == MPI_UNDEFINED) {
    fprintf(stderr,"count failed\n");
  }
  if (!*size) (*size)++;
  if ((!*message) && !((*message=malloc(*size)))) {
    fprintf(stderr,"malloc failure");
  }
  *type = status.MPI_TAG; *from = status.MPI_SOURCE;
  if (MPI_Recv(*message, *size, MPI_CHAR, *from, *type, MPI_COMM_WORLD,
	       &status) != MPI_SUCCESS) {
    fprintf(stderr, "Receive failed\n");
  }
}

void q4_msg_free(message)
char *message;
{free(message);}

  
void *q4_malloc(s)
int s;
{return malloc(s?s:s++);}

void q4_free(s)
void *s;
{free(s);}

void q4_initenv(argc, argv)
int *argc;
char **argv;
{}

void q4_create_procgroup()
{}

void q4_wait_for_end()
{}

extern int Master, Slave;

int main(int, char*[]);

int
main(argc,argv)
int argc;
char **argv;
{ 
  int id;

  if ( MPI_Init(&argc, &argv) != MPI_SUCCESS ) {
    fprintf(stderr, "Could not initialize MPI\n");
    exit(1);
  }

  id = q4_get_my_id();
  if      (id == DNAML_HOST_ID)       hostmain(argc,argv);
  else if (id == DNAML_DISPATCHER_ID) dispatchslave();
  else if (id == DNAML_MERGER_ID)     mergeslave();
  else if (id == DNAML_MASTER_ID)     { Master=1; Slave=0; slave();}
  else                                { Master=0; Slave=1; slave();}

  if (MPI_Finalize() != MPI_SUCCESS) {
    fprintf(stderr, "Could not finalize MPI\n");
    exit(1);
  }
  return 0;
}
